Node[] nodes;
Edge[] edges;
Node selectedNode = null;

float repulsionForceConstant = 10000; // ノード間の斥力のグローバル変数
float springConstant = 0.01; // ウィンドウ中心に向かうバネ定数のグローバル変数
float attractionForceConstant = 0.01; // ノード間の引力のグローバル変数

float edgeStrokeWeightMax = 8; // エッジの最大太さ
float edgeStrokeWeightMin = 1; // エッジの最小太さ
float edgeWeightMax; // エッジの重みの最大値
float edgeWeightMin; // エッジの重みの最小値

float nodeRadiusMax = 50; // ノードの最大半径
float nodeRadiusMin = 10; // ノードの最小半径
float nodeSizeMax; // ノードのサイズの最大値
float nodeSizeMin; // ノードのサイズの最小値

void setup() {
  size(800, 800);

  // CSVファイルからノードとエッジをロード
  Table nodesTable = loadTable("nodes_positive_cooccurrence_adjectives.csv", "header");
  Table edgesTable = loadTable("edges_positive_cooccurrence_adjectives.csv", "header");

  // ノードとエッジの配列を初期化
  int nodesRowCount = nodesTable.getRowCount();
  int edgesRowCount = edgesTable.getRowCount();
  nodes = new Node[nodesRowCount];
  edges = new Edge[edgesRowCount];

  // nodeSizeMaxとnodeSizeMinを極端な値で初期化
  nodeSizeMax = Float.MIN_VALUE;
  nodeSizeMin = Float.MAX_VALUE;

  // nodeSizeMaxとnodeSizeMinを決定するための最初のパス
  // ノードのサイズの最大値と最小値を決定する
  for (int i = 0; i < nodesRowCount; i++) {
    TableRow row = nodesTable.getRow(i);
    float size = row.getFloat("size");
    if (size > nodeSizeMax) {
      nodeSizeMax = size;
    }
    if (size < nodeSizeMin) {
      nodeSizeMin = size;
    }
  }

  // ノードCSVファイルからノードを作成
  // ノードごとに位置、ラベル、サイズを設定
  for (int i = 0; i < nodesRowCount; i++) {
    TableRow row = nodesTable.getRow(i);
    String label = row.getString("word");
    float size = row.getFloat("size");
    float x = random(width); // ノードの初期位置をランダムに設定
    float y = random(height);
    nodes[i] = new Node(x, y, label, size);
  }

  // edgeWeightMaxとedgeWeightMinを極端な値で初期化
  edgeWeightMax = Float.MIN_VALUE;
  edgeWeightMin = Float.MAX_VALUE;

  // エッジCSVファイルからエッジを作成
  // エッジごとにノード間のリンクを設定
  int edgeIndex = 0;
  for (int i = 0; i < edgesRowCount; i++) {
    TableRow row = edgesTable.getRow(i);
    String labelA = row.getString("word1");
    String labelB = row.getString("word2");
    float weight = row.getFloat("weight");
    
    Node a = findNodeByLabel(labelA);
    Node b = findNodeByLabel(labelB);
    
    if (a != null && b != null && a != b) {
      edges[edgeIndex++] = new Edge(a, b, weight);
      
      // edgeWeightMaxとedgeWeightMinを更新
      // エッジの重みの最大値と最小値を更新する
      if (weight > edgeWeightMax) {
        edgeWeightMax = weight;
      }
      if (weight < edgeWeightMin) {
        edgeWeightMin = weight;
      }
    }
  }
}

void draw() {
  background(255); // 背景を白にリセット

  // エッジを更新および描画
  for (Edge e : edges) {
    if (e != null) {
      e.update(); // エッジの引力を更新
      e.display(); // エッジを描画
    }
  }

  // ノードを更新および描画
  for (Node n : nodes) {
    n.update(); // ノードの位置を更新
    n.display(); // ノードを描画
  }
}

void mousePressed() {
  // マウスでノードをクリックしたときに選択する
  for (Node n : nodes) {
    if (dist(mouseX, mouseY, n.position.x, n.position.y) < n.radius) {
      selectedNode = n;
      break;
    }
  }
}

void mouseDragged() {
  // ノードが選択されている場合、そのノードをドラッグで移動
  if (selectedNode != null) {
    selectedNode.position.set(mouseX, mouseY);
  }
}

void mouseReleased() {
  // マウスを離したときにノードの選択を解除
  selectedNode = null;
}

Node findNodeByLabel(String label) {
  // ラベルを使ってノードを検索する
  for (Node n : nodes) {
    if (n.label.equals(label)) {
      return n;
    }
  }
  return null;
}

class Node {
  PVector position; // ノードの位置
  PVector velocity; // ノードの速度
  float radius; // ノードの半径
  float size; // ノードのサイズ（元データから取得）
  String label; // ノードのラベル

  Node(float x, float y, String label, float size) {
    position = new PVector(x, y);
    velocity = new PVector(0, 0);
    this.label = label;
    this.size = size;
    // ノードのサイズを半径にマッピング
    this.radius = map(size, nodeSizeMin, nodeSizeMax, nodeRadiusMin, nodeRadiusMax);
  }

  void update() {
    // シンプルな力を適用して力レイアウトを実現
    for (Node other : nodes) {
      if (other != this && selectedNode != this) {
        // 他のノードとの方向を計算
        PVector dir = PVector.sub(position, other.position);
        float distance = max(dir.mag(), 5); // 距離が小さすぎないように最低値を設定
        dir.normalize();
        // 斥力の計算（距離の2乗に反比例）
        float force = repulsionForceConstant / (distance * distance);
        dir.mult(force);
        velocity.add(dir);
      }
    }
    
    // ウィンドウ中心に向かう力を適用（フックの法則によるバネの力）
    PVector center = new PVector(width / 2, height / 2);
    PVector toCenter = PVector.sub(center, position);
    toCenter.mult(springConstant);
    velocity.add(toCenter);
    
    // 速度を制限（あまり速くなりすぎないように）
    velocity.limit(5);
    
    // 選択されていない場合に位置を更新
    if (selectedNode != this) {
      position.add(velocity);
    }
    
    // 速度を減衰させる（摩擦のような効果）
    velocity.mult(0.9);

    // ノードをウィンドウ内に保持（ウィンドウの境界を超えないように）
    position.x = constrain(position.x, radius, width - radius);
    position.y = constrain(position.y, radius, height - radius);
  }

  void display() {
    // ノードを描画
    fill(100, 150, 250); // ノードの色を設定
    noStroke();
    ellipse(position.x, position.y, radius * 2, radius * 2); // ノードを円として描画
    fill(0);
    textAlign(CENTER, CENTER);
    text(label, position.x, position.y); // ノードのラベルを表示
  }
}

class Edge {
  Node a; // エッジの一方のノード
  Node b; // エッジのもう一方のノード
  float weight; // エッジの重み

  Edge(Node a, Node b, float weight) {
    this.a = a;
    this.b = b;
    this.weight = weight;
  }

  void update() {
    // 接続されたノード間に引力を適用
    PVector dir = PVector.sub(b.position, a.position);
    float distance = max(dir.mag(), 5); // 距離が小さすぎないように最低値を設定
    float force = attractionForceConstant * (distance - 100); // 引力の強さを計算
    dir.normalize();
    dir.mult(force);
    a.velocity.add(dir); // ノードaに引力を加える
    b.velocity.sub(dir); // ノードbに反対方向の力を加える
  }

  void display() {
    // エッジを描画
    stroke(0); // 線の色を設定
    // エッジの重みに基づいて線の太さを設定
    float strokeW = map(weight, edgeWeightMin, edgeWeightMax, edgeStrokeWeightMin, edgeStrokeWeightMax);
    strokeWeight(strokeW);
    line(a.position.x, a.position.y, b.position.x, b.position.y); // エッジを線として描画
  }
}
